"""
Run all checks
"""
from cldfbench.cli_util import with_datasets
from pylexibank.cli_util import add_dataset_spec

from pylexibank.commands.check_languages import check as check_languages
from pylexibank.commands.check_lexibank import check as check_lexibank

CHECKERS = [check_languages, check_lexibank]


def register(parser):
    add_dataset_spec(parser, multiple=True)


def check(ds, args):
    for checker in CHECKERS:
        checker(ds, args, warnings=args.warnings)


def run(args):
    args.warnings = []
    with_datasets(args, check)
    if args.warnings:
        args.log.warning('{0} warnings issued'.format(len(args.warnings)))
        return 2
    args.log.info('OK')



"""
Check lexibank plumbing for lexibank datasets
"""
import functools

from cldfbench.cli_util import with_datasets

from pylexibank.cli_util import add_dataset_spec, warning


def register(parser):
    add_dataset_spec(parser, multiple=True)


def check(ds, args, warnings=None):
    warn = functools.partial(warning, args, dataset=ds, warnings=warnings)

    args.log.info('checking {0} - plumbing'.format(ds))

    # check that there's no local concepts.csv:
    if (ds.dir / 'etc' / 'concepts.csv').exists():
        warn('Dataset uses a local ./etc/concepts.csv rather than a conceptlist from Concepticon')

    # check for empty metadata fields:
    for field in ['citation', 'title', 'license']:
        if getattr(ds.metadata, field, None) in (None, ""):
            warn("Dataset has an empty '{0}' in metadata file".format(field))

    # check lexemes.csv
    if (ds.dir / 'etc' / 'lexemes.csv').exists():
        cldf = ds.cldf_reader()
        try:
            values = set(f['Value'] for f in cldf['FormTable'])
            for r in ds.lexemes:
                if r not in values:
                    warn("lexemes.csv contains un-needed conversion '{0}'".format(r))
        except KeyError:
            warn('Dataset does not seem to be a lexibank dataset - FormTable has no Value column!')


def run(args):
    with_datasets(args, check)



"""
Connect to the database with sqlite3.
"""
import subprocess

import termcolor

from pylexibank.cli_util import add_db
from pylexibank.db import Database


def register(parser):
    add_db(parser)


def run(args):
    db = str(Database(args.db).fname)
    args.log.info('connecting to {0}'.format(termcolor.colored(db, 'green')))
    subprocess.check_call(['sqlite3', db])



import lingpy
from clldutils.misc import slug
import pycldf


def wordlist2cognates(wordlist, source, expert='expert', ref='cogid'):
    """Turn a wordlist into a cognate set list, using the cldf parameters."""
    for k in wordlist:
        yield dict(
            Form_ID=wordlist[k, 'lid'],
            ID=k,
            Form=wordlist[k, 'ipa'],
            Cognateset_ID='{0}-{1}'.format(
                slug(wordlist[k, 'concept']), wordlist[k, ref]),
            Cognate_Detection_Method=expert,
            Source=source)


def _cldf2wld(dataset):
    """Make lingpy-compatible dictionary out of cldf main data."""
    forms = dataset.objects['FormTable'] \
        if hasattr(dataset, 'objects') else list(dataset['FormTable'])
    if not forms:
        raise ValueError('No forms')

    idcol = dataset['FormTable', 'id'].name if isinstance(dataset, pycldf.Dataset) else 'ID'
    header = [f for f in forms[0].keys() if f != idcol]
    D = {0: ['lid'] + [h.lower() for h in header]}
    for idx, row in enumerate(forms):
        D[idx + 1] = [row[idcol]] + [row[h] for h in header]
    return D


def _cldf2lexstat(
        dataset,
        segments='segments',
        transcription='value',
        row='parameter_id',
        col='language_id'):
    """Read LexStat object from cldf dataset."""
    D = _cldf2wld(dataset)
    return lingpy.LexStat(D, segments=segments, transcription=transcription, row=row, col=col)


def _cldf2wordlist(dataset, row='parameter_id', col='language_id'):
    """Read worldist object from cldf dataset."""
    return lingpy.Wordlist(_cldf2wld(dataset), row=row, col=col)


def iter_cognates(dataset, column='Segments', method='turchin', threshold=0.5, **kw):
    """
    Compute cognates automatically for a given dataset.

    :param dataset: Either a `LexibankWriter` instance or a `pycldf.Dataset`.
    """
    forms = dataset.objects['FormTable'] \
        if hasattr(dataset, 'objects') else list(dataset['FormTable'])

    if method == 'turchin':
        for row in forms:
            sounds = ''.join(lingpy.tokens2class(row[column], 'dolgo'))
            if sounds.startswith('V'):
                sounds = 'H' + sounds
            sounds = '-'.join([s for s in sounds if s != 'V'][:2])
            cogid = slug(row['Parameter_ID']) + '-' + sounds
            if '0' not in sounds:
                yield dict(
                    Form_ID=row['ID'],
                    Form=row['Value'],
                    Cognateset_ID=cogid,
                    Cognate_Detection_Method='CMM')

    if method in ['sca', 'lexstat']:
        try:
            lex = _cldf2lexstat(dataset)
        except ValueError:
            return
        if method == 'lexstat':
            lex.get_scorer(**kw)
        lex.cluster(method=method, threshold=threshold, ref='cogid')
        for k in lex:
            yield dict(
                Form_ID=lex[k, 'lid'],
                Form=lex[k, 'value'],
                Cognateset_ID=lex[k, 'cogid'],
                Cognate_Detection_Method=method + '-t{0:.2f}'.format(threshold))


def iter_alignments(dataset, cognate_sets, column='Segments', method='library', almkw=None):
    """
    Function computes automatic alignments and writes them to file.
    """
    if not isinstance(dataset, lingpy.basic.parser.QLCParser):
        try:
            wordlist = _cldf2wordlist(dataset)
        except ValueError:
            return
        cognates = {r['Form_ID']: r for r in cognate_sets}
        wordlist.add_entries(
            'cogid',
            'lid',
            lambda x: cognates[x]['Cognateset_ID'] if x in cognates else 0)
        alm = lingpy.Alignments(
            wordlist,
            ref='cogid',
            row='parameter_id',
            col='language_id',
            transcription='form',
            segments=column.lower())
        alm.align(method=method)
        for k in alm:
            if alm[k, 'lid'] in cognates:
                cognate = cognates[alm[k, 'lid']]
                cognate['Alignment'] = alm[k, 'alignment']
                cognate['Alignment_Method'] = method
    else:
        almkw = almkw or {}
        almkw.setdefault('ref', 'cogid')
        alm = lingpy.Alignments(dataset, **almkw)
        alm.align(method=method)

        for cognate in cognate_sets:
            idx = cognate.get('ID') or cognate['Form_ID']
            cognate['Alignment'] = alm[int(idx), 'alignment']
            cognate['Alignment_Method'] = 'SCA-' + method



import re
import logging
import unicodedata

import attr
from clldutils import text
from clldutils import misc

__all__ = ['FormSpec']
log = logging.getLogger('pylexibank')


def valid_replacements(instance, attribute, value):
    if not isinstance(value, list):
        raise ValueError('replacements must be list of pairs')
    for v in value:
        if not (isinstance(v, tuple)
                and len(v) == 2
                and isinstance(v[0], str)
                and isinstance(v[1], str)):
            raise ValueError('replacements must be list of pairs')


def valid_separators(instance, attribute, value):
    if not isinstance(value, str):
        if not isinstance(value, (list, tuple)):
            raise ValueError('separators must be an iterable of single character strings')
        for v in value:
            if (not isinstance(v, str)) or len(v) > 1:
                raise ValueError('separators must be an iterable of single character strings')


def attrib(help, **kw):
    kw['metadata'] = dict(help=help)
    return attr.ib(**kw)


@attr.s
class FormSpec(object):
    """
    Specification of the value-to-form processing in Lexibank datasets:

    The value-to-form processing is divided into two steps, implemented as methods:
    - `FormSpec.split`: Splits a string into individual form chunks.
    - `FormSpec.clean`: Normalizes a form chunk.

    These methods use the attributes of a `FormSpec` instance to configure their behaviour.
    """
    brackets = attrib(
        "Pairs of strings that should be recognized as brackets, specified as `dict` "
        "mapping opening string to closing string",
        default={"(": ")"},
        validator=attr.validators.instance_of(dict),
    )
    separators = attrib(
        "Iterable of single character tokens that should be recognized as word separator",
        default=(";", "/", ","),
        validator=valid_separators,
    )
    missing_data = attrib(
        "Iterable of strings that are used to mark missing data",
        default=('?', '-'),
        validator=attr.validators.instance_of((tuple, list)),
    )
    strip_inside_brackets = attrib(
        "Flag signaling whether to strip content in brackets "
        "(**and** strip leading and trailing whitespace)",
        default=True,
        validator=attr.validators.instance_of(bool))
    replacements = attrib(
        "List of pairs (`source`, `target`) used to replace occurrences of `source` in forms"
        "with `target` (before stripping content in brackets)",
        default=attr.Factory(list),
        validator=valid_replacements)
    first_form_only = attrib(
        "Flag signaling whether at most one form should be returned from `split` - "
        "effectively ignoring any spelling variants, etc.",
        default=False,
        validator=attr.validators.instance_of(bool),
    )
    normalize_whitespace = attrib(
        "Flag signaling whether to normalize whitespace - stripping leading and trailing "
        "whitespace and collapsing multi-character whitespace to single spaces",
        default=True,
        validator=attr.validators.instance_of(bool),
    )
    normalize_unicode = attrib(
        "UNICODE normalization form to use for input of `split` (`None`, 'NFD' or 'NFC')",
        default=None,
        validator=attr.validators.in_(['NFD', 'NFC', None]),
    )

    def as_markdown(self, dataset=None):
        """
        :return: Description of `FormSpec` in markdown.
        """
        res = ['## Specification of form manipulation\n']
        res.extend([line.strip() for line in self.__class__.__doc__.splitlines()])
        for field in attr.fields(self.__class__):
            res.extend([
                '- `{0}`: `{1}`'.format(field.name, getattr(self, field.name)),
                '  {0}'.format(field.metadata['help'])
            ])
        if dataset:
            if dataset.lexemes:
                res.append("""
### Replacement of invalid lexemes

Source lexemes may be impossible to interpret correctly. {0} such lexemes are listed
in [`etc/lexemes.csv`](etc/lexemes.csv) and replaced as specified in this file.
""".format(len(dataset.lexemes)))

            if dataset.segments:
                res.append("""
### Replacement of invalid segmentation

Segments provided in the source data may not be valid according to CLTS.
{0} such segments are listed in [`etc/segments.csv`](etc/segments.csv) and replaced
as specified in this file.
""".format(len(dataset.segments)))
        return '\n'.join(res)

    def clean(self, form, item=None):
        """
        Called when a row is added to a CLDF dataset.

        :param form:
        :return: None to skip the form, or the cleaned form as string.
        """
        if form not in self.missing_data:
            if self.normalize_whitespace:
                form = re.sub(r'\s+', ' ', form.strip())
            for source, target in self.replacements:
                form = form.replace(source, target)
            if self.strip_inside_brackets:
                return text.strip_brackets(form, brackets=self.brackets)
            return form

    def split(self, item, value, lexemes=None):
        lexemes = lexemes or {}
        if value in lexemes:
            log.debug('overriding via lexemes.csv: %r -> %r' % (value, lexemes[value]))
            value = lexemes[value]
        if self.normalize_unicode:
            value = unicodedata.normalize(self.normalize_unicode, value)
        res = misc.nfilter(
            self.clean(form, item=item)
            for form in text.split_text_with_context(
                value, separators=self.separators, brackets=self.brackets))
        if self.first_form_only:
            return res[:1]
        return res



import pkg_resources

import attr
from cldfbench.metadata import Metadata

__all__ = ['LexibankMetadata']

version = pkg_resources.get_distribution('pylexibank').version


@attr.s
class LexibankMetadata(Metadata):
    aboutUrl = attr.ib(default=None)
    conceptlist = attr.ib(
        default=[],
        converter=lambda s: [] if not s else (s if isinstance(s, list) else [s]))
    lingpy_schema = attr.ib(default=None)
    derived_from = attr.ib(default=None)
    related = attr.ib(default=None)
    source = attr.ib(default=None)
    patron = attr.ib(default=None)
    version = attr.ib(default=version)

    def common_props(self):
        res = super().common_props()
        res.update({
            "dc:format": [
                "http://concepticon.clld.org/contributions/{0}".format(cl)
                for cl in self.conceptlist],
            "dc:isVersionOf": "http://lexibank.clld.org/contributions/{0}".format(
                self.derived_from) if self.derived_from else None,
            "dc:related": self.related,
            "aboutUrl": self.aboutUrl
        })
        return res

    def markdown(self):
        lines = [super().markdown(), '']

        if self.related:
            lines.append('See also %s\n' % self.related)

        if self.conceptlist:
            lines.append('Conceptlists in Concepticon:')
            lines.extend([
                '- [{0}](https://concepticon.clld.org/contributions/{0})'.format(cl)
                for cl in self.conceptlist])
            lines.append('')
        return '\n'.join(lines)



import pathlib

from cldfbench import cli_util
from termcolor import colored

from pylexibank import ENTRY_POINT
from pylexibank.db import Database


def add_dataset_spec(parser, **kw):
    kw.setdefault('ep', ENTRY_POINT)
    return cli_util.add_dataset_spec(parser, **kw)


def add_db(parser):
    parser.add_argument(
        '--db',
        help='path to SQLite db file',
        type=pathlib.Path,
        default=pathlib.Path.cwd() / 'lexibank.sqlite')


def get_db(args):
    db = Database(args.db)
    db.create(exists_ok=True)
    return db


def add_catalogs(parser, with_clts=False):
    cli_util.add_catalog_spec(parser, 'glottolog')
    cli_util.add_catalog_spec(parser, 'concepticon')
    if with_clts:
        cli_util.add_catalog_spec(parser, 'clts')


def warning(args, msg, dataset=None, warnings=None):
    if dataset:
        msg = '{0}: {1}'.format(colored(dataset.id, 'blue', attrs=['bold']), msg)
    args.log.warning(msg)
    if warnings is not None:
        warnings.append(msg)



import re

from bs4 import BeautifulSoup
import bs4
from csvw.dsv import UnicodeWriter

from pylexibank.dataset import Dataset
from pylexibank.util import pb, getEvoBibAsBibtex

SOURCE = 'Starostin2011'


class TOB(Dataset):
    name = None
    dset = None
    pages = 1
    lexemes = {}

    def _url(self, page):
        return 'http://starling.rinet.ru/cgi-bin/response.cgi?' + \
            'root=new100&morpho=0&basename=new100' + \
            r'\{0}\{1}&first={2}'.format(self.dset, self.name, page)

    def cmd_download(self, args):
        # download source
        self.raw_dir.write('sources.bib', getEvoBibAsBibtex(SOURCE, **vars(args)))

        # download data
        all_records = []
        for i in pb(list(range(1, 20 * self.pages + 1, 20))):
            with self.raw_dir.temp_download(
                    self._url(i), 'file-{0}'.format(i), log=args.log) as fname:
                soup = BeautifulSoup(fname.open(encoding='utf8').read(), 'html.parser')
                for record in soup.findAll(name='div', attrs={"class": "results_record"}):
                    if isinstance(record, bs4.element.Tag):
                        children = list(record.children)
                        number = children[0].findAll('span')[1].text.strip()
                        concept = children[1].findAll('span')[1].text
                        for child in children[2:]:
                            if isinstance(child, bs4.element.Tag):
                                dpoints = child.findAll('span')
                                if len(dpoints) >= 3:
                                    lname = dpoints[1].text
                                    glottolog = re.findall(
                                        'Glottolog: (........)', str(dpoints[1]))[0]
                                    entry = dpoints[2].text
                                    cogid = list(child.children)[4].text.strip()
                                    all_records.append(
                                        (number, concept, lname, glottolog, entry, cogid))
        with UnicodeWriter(self.raw_dir / 'output.csv') as f:
            f.writerows(all_records)

    def cmd_makecldf(self, args):
        ds = args.writer
        ds.add_sources()
        ds.add_concepts(id_factory=lambda c: c.number)
        for cid, concept, lid, gc, form, cogid in pb(self.raw_dir.read_csv('output.csv')):
            ds.add_language(ID=lid.replace(' ', '_'), Name=lid, Glottocode=gc)
            for row in ds.add_lexemes(
                Language_ID=lid.replace(' ', '_'),
                Parameter_ID=cid,
                Value=form,
                Source=[SOURCE],
                Cognacy=concept + '-' + cogid
            ):
                ds.add_cognate(
                    lexeme=row,
                    Cognateset_ID=cogid,
                    Source=[SOURCE])



from setuptools import setup, find_packages


setup(
    name='pylexibank',
    version='2.3.0',
    author='Robert Forkel',
    author_email='forkel@shh.mpg.de',
    description='Python library implementing the lexibank workbench',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    keywords='',
    license='Apache 2.0',
    url='https://github.com/lexibank/pylexibank',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'cldfbench.commands': [
            'lexibank=pylexibank.commands',
        ],
        'cldfbench.scaffold': [
            'lexibank_simple=pylexibank.scaffold:LexibankTemplate',
            'lexibank_combined=pylexibank.scaffold:LexibankCombinedTemplate',
        ],
    },
    platforms='any',
    python_requires='>=3.5',
    install_requires=[
        'segments>=2.1.1',
        'cldfbench[excel]>=1.0',
        'csvw>=1.5.6',
        'clldutils>=2.8.0',
        'pycldf>=1.7.0',
        'attrs>=18.1.0',
        'pyglottolog>=2.0',
        'pyconcepticon>=2.1.0',
        'pyclts>=2.0',
        'lingpy>=2.6.5',
        'appdirs',
        'requests',
        'termcolor',
        'gitpython',
        'tqdm',
        'bs4',
    ],
    extras_require={
        'dev': ['flake8', 'wheel', 'twine'],
        'test': [
            'mock',
            'pytest>=3.6',
            'pytest-mock',
            'pytest-cov',
            'coverage>=4.2',
        ],
    },
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
)



import attr

__all__ = ['Language', 'Lexeme', 'Concept', 'Cognate']


def non_empty(_, attribute, value):
    if not value:
        raise ValueError('{0} must be non-empty'.format(attribute))


class FieldnamesMixin(object):
    @classmethod
    def fieldnames(cls):
        return [f.name for f in attr.fields(cls)]


@attr.s(hash=False)
class Language(FieldnamesMixin):
    ID = attr.ib(default='', hash=True)
    Name = attr.ib(default=None)
    ISO639P3code = attr.ib(default=None)
    Glottocode = attr.ib(default=None)
    Macroarea = attr.ib(default=None)
    Latitude = attr.ib(default=None)
    Longitude = attr.ib(default=None)

    Glottolog_Name = attr.ib(default=None)
    Family = attr.ib(default=None)

    @classmethod
    def __cldf_table__(cls):
        return 'LanguageTable'


@attr.s(hash=False)
class Concept(FieldnamesMixin):
    ID = attr.ib(default='', hash=True)
    Name = attr.ib(default='')
    Concepticon_ID = attr.ib(default=None)
    Concepticon_Gloss = attr.ib(default=None)

    @classmethod
    def __cldf_table__(cls):
        return 'ParameterTable'


@attr.s
class Lexeme(FieldnamesMixin):
    """
    Raw lexical data item as it can be pulled out of the original datasets.

    This is the basis for creating rows in CLDF representations of the data by
    - splitting the lexical item into forms
    - cleaning the forms
    - potentially tokenizing the form
    """
    ID = attr.ib()
    Form = attr.ib()
    Value = attr.ib(validator=non_empty)  # the lexical item
    Language_ID = attr.ib(validator=non_empty)
    Parameter_ID = attr.ib(validator=non_empty)
    Local_ID = attr.ib(default=None)  # local ID of a lexeme in the source dataset
    Segments = attr.ib(
        default=attr.Factory(list),
        validator=attr.validators.instance_of(list))
    Source = attr.ib(
        default=attr.Factory(list),
        validator=attr.validators.instance_of(list),
        converter=lambda v: [v] if isinstance(v, str) else v)
    Comment = attr.ib(default=None)
    Cognacy = attr.ib(default=None)
    Loan = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(bool)))

    @classmethod
    def __cldf_table__(cls):
        return 'FormTable'


@attr.s
class Cognate(FieldnamesMixin):
    ID = attr.ib(default=None)
    Form_ID = attr.ib(default=None)
    Form = attr.ib(default=None)
    Cognateset_ID = attr.ib(default=None)
    Doubt = attr.ib(
        default=False,
        converter=lambda v: v if isinstance(v, bool) else eval(v))
    Cognate_Detection_Method = attr.ib(default='expert')
    Source = attr.ib(
        default=attr.Factory(list),
        validator=attr.validators.instance_of(list),
        converter=lambda v: [v] if isinstance(v, str) else v)
    Alignment = attr.ib(
        default=None,
        converter=lambda v: v if isinstance(v, list) or v is None else v.split())
    Alignment_Method = attr.ib(default=None)
    Alignment_Source = attr.ib(default=None)

    @classmethod
    def __cldf_table__(cls):
        return 'CognateTable'
